from urllib.request import urlopen
import pandas as pd

def loaddata(dataframe, variable_lst):
    df = dataframe
    dep_var_name = variable_lst[0]
    endog_var_name = variable_lst[1]
    exog_var_list = variable_lst[2]
    iv_var_list = variable_lst[3]

    # define the dependent variable (Y) and the endogenous right-hand side variable (X)
    dep_var = df[dep_var_name].to_numpy()
    endog_var = df[endog_var_name].to_numpy()

    # define exogenous included variables
    exog_var = df[exog_var_list].to_numpy()

    # define all exogenous variables
    z_var_list = exog_var_list + iv_var_list
    z_var = df[z_var_list].to_numpy()

    dim_z = len(z_var_list)

    return dep_var, endog_var, exog_var, z_var, dim_z


from .eval_bspline_basis import eval_bspline_basis

def dev_lambda(exog_var, beta, option, degree, vhat):
    # dim_endo_exo = total length of endogenous variable and exogenous variables
    dim_endo_exo = 1 + exog_var.shape[1]
    derivative_lambda = 0

    # option 0: polynomial series
    if option == 0:
        # `beta` has length `dim_endo_exo` + `degree`
        # we need the last `degree` terms in `beta`, so we skip the first `dim_endo_exo` terms of `beta`
        # Note that the derivative of vhat^i is i*vhat^(i-1)
        # Thus, each derivative term of vhat, vhat^2, ..., vhat^degree is
        # 1, 2*vhat, 3*vhat^2, ..., degree*vhat^(degree-1)
        # Each term corresponds to beta[dim_endo_exo], ..., beta[dim_endo_exo+degree-1]
        for i in range(1, degree+1):
            derivative_lambda += i*(vhat**(i-1))*beta.value[dim_endo_exo+i-1]

    # option 1: B-splines
    if option == 1:
        # der = 1 for derivative
        derivative_lambda = 0
        all_derivative = eval_bspline_basis(x=vhat, df=2*degree-1, degree=degree, der=1, include_intercept=False)
        num_columns = all_derivative.shape[1]

        # `beta` has length `dim_endo_exo` + 2*`degree`-1
        # `all_derivative` has 2*`degree`-1 columns (`num_columns` = 2*`degree`-1)
        # we need the last `num_columns` terms in `beta`, so we skip the first `dim_endo_exo` terms of `beta`
        # Each column in `derivative_lambda` corresponds to beta[dim_endo_exo], ..., beta[dim_endo_exo+num_columns-1]
        for i in range(num_columns):
            derivative_lambda += all_derivative[:, i] * beta.value[dim_endo_exo+i]

    return derivative_lambda


import numpy as np
import cvxpy as cp
from .eval_bspline_basis import eval_bspline_basis

def second_stage_reg(dep_var, endog_var, exog_var, tau_second_stage, degree, option, vhat):
    # create w_hat and vhat for the second stage regression
    # option 0: polynomial series
    if option == 0:
        w_hat = [endog_var, exog_var]
        for j in range(1, degree + 1):
            w_hat.append(np.power(vhat, j))
        w_hat = np.column_stack(w_hat)

    # option 1: B-splines
    if option == 1:
        vhat_bs_hand = eval_bspline_basis(x=vhat, df=2 * degree - 1, degree=degree, der=0, include_intercept=False)
        w_hat = np.column_stack([endog_var, exog_var, vhat_bs_hand])

    # Create optimization variables and a parameter
    dim_w = w_hat.shape[1]
    beta = cp.Variable(dim_w)
    alpha = cp.Variable()
    tau = cp.Parameter()

    # Create the residual and error term of quantile regression model
    fit = w_hat@beta + alpha
    resid = dep_var - fit
    error = cp.sum(0.5*cp.abs(resid) + (tau - 0.5)*resid)

    # Form the problem
    prob = cp.Problem(cp.Minimize(error))
    tau.value = tau_second_stage
    prob.solve()

    second_stage_residual = resid.value

    return w_hat, beta, second_stage_residual


from .loaddata import loaddata
from .first_reg import first_stage_reg
from .second_reg import second_stage_reg
from .dev_lambda import dev_lambda
from .asympt_variance import asympt_variance

# data = [dataframe, variable_lst]
def quantregCF(option=1, degree=3, tau_first_stage=0.5, tau_second_stage=0.5, data=[]):
    # load the dataset
    dep_var, endog_var, exog_var, z_var, dim_z = loaddata(data[0], data[1])

    # regressions
    vhat = first_stage_reg(endog_var, z_var, dim_z, tau_first_stage)
    w_hat, beta, second_stage_res = second_stage_reg(dep_var, endog_var, exog_var, tau_second_stage, degree, option, vhat)
    beta_lst = [beta.value[i] for i in range(beta.shape[0])]
    derivative_lambda = dev_lambda(exog_var, beta, option, degree, vhat)
    se = asympt_variance(exog_var, z_var, tau_second_stage, w_hat, vhat, second_stage_res, derivative_lambda)

    return beta_lst, se


import cvxpy as cp

def first_stage_reg(endog_var, z_var, dim_z, tau_first_stage):
    # Create optimization variables and a parameter
    beta = cp.Variable(dim_z)
    alpha = cp.Variable()
    tau = cp.Parameter()
    tau.value = tau_first_stage

    # Create the residual and error term of quatile regression model
    fit = z_var@beta + alpha
    resid = endog_var - fit
    error = cp.sum(0.5*cp.abs(resid) + (tau - 0.5)*resid)

    # Form the problem
    prob = cp.Problem(cp.Minimize(error))
    prob.solve()

    # Estimate the value of V to be used in the second stage
    vhat = resid.value
    return vhat



import numpy as np
from .kernel import kl_normal

def asympt_variance(exog_var, z_var, tau_second_stage, w_hat, vhat, second_stage_residual, derivative_lambda):
    # estimate K
    tau = tau_second_stage
    n = len(second_stage_residual)
    bdw = np.std(second_stage_residual)*np.power(n, -1/5)
    kl_weight = kl_normal(second_stage_residual/bdw)/bdw
    Khat = np.diag(kl_weight)

    # estimate Phi
    Phat = np.column_stack([w_hat, np.tile(1, (len(w_hat), 1))])
    Phihat = Phat.T@Khat@Phat/n

    # estimate Sigma
    Sigmahat = tau*(1-tau)*Phat.T@Phat/n

    # estimate Gamma
    tmp = kl_weight*derivative_lambda
    KLdhat = np.diag(tmp)
    Zhat = np.column_stack([z_var, np.tile(1, (len(z_var), 1))])
    Gammahat = Phat.T@KLdhat@Zhat/n

    # estimate Sigma_first_stage
    bdw = np.std(vhat)*np.power(n, -1/5)
    kl_weight = kl_normal(vhat/bdw)/bdw
    Khat = np.diag(kl_weight)
    tmp0 = Zhat.T@Khat@Zhat/n
    tmp1 = tau*(1-tau)*Zhat.T@Zhat/n
    Sigma_first_stage = np.linalg.inv(tmp0)@tmp1@np.linalg.inv(tmp0)

    # estimate Omega
    # first, specify A
    dim_z1 = len(exog_var[0])
    dim_w = len(Phat[0])
    Amat1 = np.eye(1+dim_z1)
    Amat2 = np.zeros((1+dim_z1, dim_w-1-dim_z1))
    Amat = np.column_stack([Amat1, Amat2])

    # then, estimate Omega
    Omega1 = Amat@np.linalg.inv(Phihat)
    Omega2 = Sigmahat + Gammahat@Sigma_first_stage@Gammahat.T
    Omega = Omega1@Omega2@Omega1.T

    # compute the standard error
    avar = np.diag(Omega/n)
    se = np.sqrt(avar)

    return se




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
        name="quantregCF",
        version="1.0.0",
        packages=["quantregCF"],
        description="The control function approach for quantile regression models",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/quantregCF/quantregCF-py",
        author="Sokbae (Simon) Lee, Suwijak (Max) Mongkalakorn",
        author_email="sl3841@columbia.edu, sm4791@columbia.edu",
        license="MIT",
        classifiers=[
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
        ],
        install_requires=[
            "numpy",
            "pandas",
            "scipy",
            "cvxpy"
            ],
        package_dir = {"": "src"}
        )



import numpy as np
from scipy.interpolate import splev

def R_compat_quantile(x, probs):
    #return np.percentile(x, 100 * np.asarray(probs))
    probs = np.asarray(probs)
    quantiles = np.asarray([np.percentile(x, 100 * prob) for prob in probs.ravel(order="C")])
    return quantiles.reshape(probs.shape, order="C")

def eval_bspline_basis(x, df, degree, der, include_intercept):
    # Note: the order of a spline is the same as its degree + 1.
    # Note: there are (len(knots) - order) basis functions.
    order = degree + 1
    n_inner_knots = df - order
    if not include_intercept:
        n_inner_knots += 1
    knot_quantiles = np.linspace(0, 1, n_inner_knots + 2)[1:-1]
    inner_knots = R_compat_quantile(x, knot_quantiles)
    knots = np.concatenate(([np.min(x), np.max(x)] * order, inner_knots))
    knots.sort()
    n_bases = len(knots) - (degree + 1)
    basis = np.empty((x.shape[0], n_bases), dtype=float)
    for i in range(n_bases):
        coefs = np.zeros((n_bases,))
        coefs[i] = 1
        basis[:, i] = splev(x, (knots, coefs, degree), der)
    if not include_intercept:
            basis = basis[:, 1:]
    return basis
